#include "audio_in.h"

Audio_In::Audio_In(QObject *parent) :
        QObject(parent)
      ,audio_out_stream(NULL)
{
        audio_format.setSampleRate(44100); //can be one of 22050, 44100, 11025, 48000, 8000 (20111004) - must be 44100 for streaming (flv container) (20130312)
        audio_format.setChannelCount(2);
        audio_format.setSampleSize(16);
        audio_format.setCodec("audio/pcm");
        audio_format.setByteOrder(QAudioFormat::LittleEndian);
        audio_format.setSampleType(QAudioFormat::SignedInt);

        level_indicator_throttle.break_time_ms = 20;
}

void Audio_In::enumerate_devices() {
        QStringList input_device_names, output_device_names;

        foreach (const QAudioDeviceInfo &deviceInfo, QAudioDeviceInfo::availableDevices(QAudio::AudioInput)) {
                input_device_names << deviceInfo.deviceName();
        }

        foreach (const QAudioDeviceInfo &deviceInfo, QAudioDeviceInfo::availableDevices(QAudio::AudioOutput)) {
                output_device_names << deviceInfo.deviceName();
        }

        emit found_devices(input_device_names, output_device_names);
}

void Audio_In::audio_start(QString input_device_name, QString output_device_name, bool new_playthrough) {
        qDebug() << this << "audio_start()" << input_device_name << output_device_name << new_playthrough;
        if (audio_out_stream) {
                audio_out_stream->stop();
                delete audio_out_stream;
                audio_out_stream = NULL;
        }
        playthrough = new_playthrough;

        QAudioDeviceInfo input_device_info = QAudioDeviceInfo::defaultInputDevice();
        QAudioDeviceInfo output_device_info = QAudioDeviceInfo::defaultOutputDevice();

        QStringList desired_input_device_names, desired_output_device_names;
        desired_input_device_names
//                #ifdef Q_OS_MAC
//                        << "Soundflower (2ch)"
//                   #endif
//                   #ifdef Q_OS_WIN32
//                        << "Stereo Mix"
//                   #endif
                        << input_device_name
                           ;
//#ifdef Q_OS_MAC
        desired_output_device_names
//                        << "Built-in Output"
//                        << "Internal Speakers"
                        << output_device_name
                           ;
//#endif
        //windows - the default is fine

        bool found_desired_input_device_name = false;
        bool found_desired_output_device_name = false;

        foreach (const QAudioDeviceInfo &deviceInfo, QAudioDeviceInfo::availableDevices(QAudio::AudioInput)) {
                foreach (QString desired_input_device_name, desired_input_device_names) {
                        if (deviceInfo.deviceName() == desired_input_device_name || deviceInfo.deviceName().contains(desired_input_device_name)) {
                                input_device_info = deviceInfo;
                                found_desired_input_device_name = true;
                                break;
                        }// else qDebug() << "rejecting input device" << deviceInfo.deviceName();
                }
        }
        if (!found_desired_input_device_name) {
                qDebug() << this << "unable to find desired input device";
        }

        foreach (const QAudioDeviceInfo &deviceInfo, QAudioDeviceInfo::availableDevices(QAudio::AudioOutput)) {
                foreach (QString desired_output_device_name, desired_output_device_names) {
                        if (deviceInfo.deviceName() == desired_output_device_name || deviceInfo.deviceName().contains(desired_output_device_name)) {
                                output_device_info = deviceInfo;
                                found_desired_output_device_name = true;
                                break;
                        }// else qDebug() << "rejecting output device" << deviceInfo.deviceName();
                }
        }
        if (!found_desired_output_device_name) {
                qDebug() << this << "unable to find desired output device";
        }

        if (!input_device_info.isFormatSupported(audio_format)) {
                qDebug() << this << "audio format not supported by input device";
                return;
        }
        if (!output_device_info.isFormatSupported(audio_format)) {
                qDebug() << this << "audio format not supported by output device";
                return;
        }

        qDebug() << this << "using input device" << input_device_info.deviceName() << "output device" << output_device_info.deviceName();

        audio_in_stream = new QAudioInput(input_device_info, audio_format);
        //this was no longer needed in qt 5.1 (20131020)
//#ifdef Q_OS_MAC
//        audio_in_stream->setBufferSize(1024 * 1024); //a large enough buffer is important to prevent fatal overruns
//#endif

        audio_device = audio_in_stream->start();
        connect(audio_device, SIGNAL(readyRead()), this, SLOT(audio_read()));

        if (playthrough) {
                audio_out_stream = new QAudioOutput(output_device_info, audio_format, this);
                audio_out_buffer = audio_out_stream->start();
        }
}

void Audio_In::audio_read() {
        QByteArray new_data = audio_device->readAll();
        if (!new_data.size()) return; //why does this happen? (20111006)
        if (playthrough) {
                audio_out_buffer->write(new_data);
        }
        if (new_data.size() >= 2 && !level_indicator_throttle.busy()) {
                qint16 *first_sample_level = (qint16 *)new_data.data();
                emit read_audio_level((qAbs(*first_sample_level) / (double)std::numeric_limits<qint16>::max()) * 100.0);
                level_indicator_throttle.done();
        }
        emit broadcast_audio_data(new_data);
}
