#include "kumari.h"

Kumari::Kumari(QWidget *parent)
        : QMainWindow(parent)
        ,currently_streaming(false)
        ,currently_recording(false)
        ,profiling(false)
        ,audio_finished_populating(false)
{
        setWindowTitle("Kumari");
        setWindowIcon(QIcon(":/Kumari.png"));
        QCoreApplication::setOrganizationName("Speed Demos Archive");
        QCoreApplication::setOrganizationDomain("speeddemosarchive.com");
        QCoreApplication::setApplicationName("Kumari");
        QCoreApplication::setApplicationVersion(QString("%1").arg(KUMARI_VERSION));
        settings = new QSettings;

        presets
                        << "ultrafast"
                        << "superfast"
                        << "veryfast"
                        << "faster"
                        << "fast"
                        << "medium"
                        << "slow"
                        << "slower"
                        << "veryslow"
                        << "placebo"
                           ;

        broadcaster.moveToThread(&broadcaster_thread);
        broadcaster.mutex = &ffmpeg_mutex;

        connect(this, SIGNAL(broadcaster_start(QString,int,int,double,QString,int,int,int,int)),
                &broadcaster, SLOT(start(QString,int,int,double,QString,int,int,int,int)));

        connect(this, SIGNAL(broadcaster_stop()),
                &broadcaster, SLOT(stop()));

        connect(this, SIGNAL(encode_frame(QImage)),
                &broadcaster, SLOT(add_video_frame(QImage)));

        connect(this, SIGNAL(frame_ready(QImage)),
                &broadcaster, SLOT(add_video_frame(QImage)));

        connect(&audio_in, SIGNAL(broadcast_audio_data(QByteArray)),
                &broadcaster, SLOT(add_audio_data(QByteArray)));

        connect(&broadcaster, SIGNAL(error_with_string(QString)),
                this, SLOT(error_dialog_slot(QString)));



        audio_in.moveToThread(&audio_in_thread);
        connect(&audio_in_thread, SIGNAL(started()),
                &audio_in, SLOT(enumerate_devices()));
        connect(&audio_in, SIGNAL(found_devices(QStringList,QStringList)),
                this, SLOT(populate_audio_devices(QStringList,QStringList)));
        connect(this, SIGNAL(audio_start(QString,QString,bool)),
                &audio_in, SLOT(audio_start(QString,QString,bool)));

        connect(&audio_in, SIGNAL(read_audio_level(int)),
                &level_indicator, SLOT(display_level(int)));


        lossless_recorder.moveToThread(&recorder_thread);
        lossless_recorder.mutex = &ffmpeg_mutex;
        connect(this, SIGNAL(frame_ready(QImage)),
                &lossless_recorder, SLOT(add_video_frame(QImage)));
        connect(&audio_in, SIGNAL(broadcast_audio_data(QByteArray)),
                &lossless_recorder, SLOT(add_audio_data(QByteArray)));
        connect(this, SIGNAL(recording_start(QString,int,int,double,QString,int,int)),
                &lossless_recorder, SLOT(start(QString,int,int,double,QString,int,int)));
        connect(this, SIGNAL(recording_stop()),
                &lossless_recorder, SLOT(stop()));


#ifdef Q_OS_MAC
        QMenuBar *menu_bar = new QMenuBar(0);
#else
        QMenuBar *menu_bar = menuBar();
#endif


        stream_menu = menu_bar->addMenu(tr("&Stream"));
        set_streaming_url_act = new QAction(tr("Set streaming &URL ..."), this);
        connect(set_streaming_url_act, SIGNAL(triggered()), this, SLOT(set_streaming_url()));
        stream_menu->addAction(set_streaming_url_act);


        record_menu = menu_bar->addMenu(tr("&Record"));
        set_recording_filepath_act = new QAction(tr("Set recording &file ..."), this);
        connect(set_recording_filepath_act, SIGNAL(triggered()), this, SLOT(set_recording_filepath()));
        record_menu->addAction(set_recording_filepath_act);


        help_menu = menu_bar->addMenu(tr("&Help"));

        about_act = new QAction(tr("&About Kumari"), this);
        connect(about_act, SIGNAL(triggered()), this, SLOT(about()));
        about_act->setMenuRole(QAction::AboutRole);

        about_qt_act = new QAction(tr("About &Qt"), this);
        connect(about_qt_act, SIGNAL(triggered()), qApp, SLOT(aboutQt()));
        about_qt_act->setMenuRole(QAction::AboutQtRole);

        help_menu->addAction(about_act);
        help_menu->addAction(about_qt_act);


        actions
                        << set_streaming_url_act
                        << set_recording_filepath_act
                           ;



        x_spinbox = make_spinbox();
        y_spinbox = make_spinbox();
        width_spinbox = make_spinbox();
        height_spinbox = make_spinbox();
        width_spinbox->setSingleStep(4);
        height_spinbox->setSingleStep(2);
        framerate_spinbox = make_spinbox();
        connect(framerate_spinbox, SIGNAL(valueChanged(int)), this, SLOT(fps_changed(int)));

        x_spinbox->setValue(settings->value("x", 0).toInt());
        y_spinbox->setValue(settings->value("y", 0).toInt());
        width_spinbox->setValue(settings->value("width", 640).toInt());
        height_spinbox->setValue(settings->value("height", 480).toInt());
        framerate_spinbox->setValue(settings->value("framerate", 60).toInt());


        audio_in_combobox = new QComboBox;
        audio_out_combobox = new QComboBox;

        audio_out_combobox->setDisabled(true);

        audio_in_combobox->setSizeAdjustPolicy(QComboBox::AdjustToContents);
        audio_out_combobox->setSizeAdjustPolicy(QComboBox::AdjustToContents);

        playthrough_checkbox = new QCheckBox(tr("playthrough"));
        connect(audio_in_combobox, SIGNAL(currentIndexChanged(QString)),
                this, SLOT(audio_device_changed(QString)));
        connect(audio_out_combobox, SIGNAL(currentIndexChanged(QString)),
                this, SLOT(audio_device_changed(QString)));
        connect(playthrough_checkbox, SIGNAL(toggled(bool)),
                this, SLOT(audio_playthrough_changed(bool)));


        stream_checkbox = new QCheckBox(tr("Stream"));
        record_checkbox = new QCheckBox(tr("Record"));
        stream_checkbox->setChecked(settings->value("stream_checkbox_checked").toBool());
        record_checkbox->setChecked(settings->value("record_checkbox_checked").toBool());



        video_bitrate_spinbox = make_spinbox();
        video_bitrate_spinbox->setValue(settings->value("video_bitrate_kbit", 512).toInt());
        audio_bitrate_spinbox = make_spinbox();
        audio_bitrate_spinbox->setValue(settings->value("audio_bitrate_kbit", 128).toInt());

        stream_slider = make_slider();
        stream_slider->setValue(settings->value("stream_preset_number", 5).toInt());
        record_slider = make_slider();
        record_slider->setValue(settings->value("record_preset_number", 5).toInt());



        widgets_to_vate
                        << x_spinbox
                        << y_spinbox
                        << width_spinbox
                        << height_spinbox
                        << framerate_spinbox
                        << video_bitrate_spinbox
                        << audio_bitrate_spinbox

                        << audio_in_combobox
                        << audio_out_combobox
                        << playthrough_checkbox

                        << stream_checkbox
                        << record_checkbox

                        << stream_slider
                        << record_slider
                           ;


        start_stop_button = new QPushButton(tr("Start"));
        connect(start_stop_button, SIGNAL(clicked()), this, SLOT(start()));

        fps_label = new QLabel("00.0");

        QHBoxLayout *horizontal_layout_1 = new QHBoxLayout;
        QHBoxLayout *horizontal_layout_2 = new QHBoxLayout;
        QHBoxLayout *horizontal_layout_3 = new QHBoxLayout;
        QHBoxLayout *horizontal_layout_4 = new QHBoxLayout;
        horizontal_layout_1->addWidget(new QLabel("x"));
        horizontal_layout_1->addWidget(x_spinbox);
        horizontal_layout_1->addWidget(new QLabel("y"));
        horizontal_layout_1->addWidget(y_spinbox);
        horizontal_layout_1->addWidget(new QLabel("w"));
        horizontal_layout_1->addWidget(width_spinbox);
        horizontal_layout_1->addWidget(new QLabel("h"));
        horizontal_layout_1->addWidget(height_spinbox);
        horizontal_layout_1->addWidget(new QLabel("fps"));
        horizontal_layout_1->addWidget(framerate_spinbox);
        horizontal_layout_1->addWidget(fps_label, 1);
        horizontal_layout_1->addWidget(start_stop_button);


        horizontal_layout_2->addWidget(new QLabel(tr("Audio in:")));
        horizontal_layout_2->addWidget(audio_in_combobox);
        horizontal_layout_2->addWidget(playthrough_checkbox);
        horizontal_layout_2->addWidget(audio_out_combobox);
        horizontal_layout_2->addWidget(&level_indicator);

        horizontal_layout_2->addStretch(1);


        int spacing_size = 10;
        horizontal_layout_3->addWidget(stream_checkbox);

        horizontal_layout_3->addSpacing(spacing_size);
        horizontal_layout_3->addWidget(new QLabel(tr("speed")));
        horizontal_layout_3->addWidget(stream_slider);
        horizontal_layout_3->addWidget(new QLabel(tr("quality")));
        horizontal_layout_3->addSpacing(spacing_size);

        horizontal_layout_3->addWidget(new QLabel(tr("video")));
        horizontal_layout_3->addWidget(video_bitrate_spinbox);
        horizontal_layout_3->addWidget(new QLabel(tr("Kbps")));

        horizontal_layout_3->addSpacing(spacing_size);

        horizontal_layout_3->addWidget(new QLabel(tr("audio")));
        horizontal_layout_3->addWidget(audio_bitrate_spinbox);
        horizontal_layout_3->addWidget(new QLabel(tr("Kbps")));

        horizontal_layout_3->addStretch(1);


        horizontal_layout_4->addWidget(record_checkbox);
        horizontal_layout_4->addSpacing(spacing_size);
        horizontal_layout_4->addWidget(new QLabel(tr("speed")));
        horizontal_layout_4->addWidget(record_slider);
        horizontal_layout_4->addWidget(new QLabel(tr("smaller files")));

        horizontal_layout_4->addStretch(1);


        vertical_layout = new QVBoxLayout;
        vertical_layout->addLayout(horizontal_layout_1);
        vertical_layout->addLayout(horizontal_layout_2);
        vertical_layout->addLayout(horizontal_layout_3);
        vertical_layout->addLayout(horizontal_layout_4);
        display_vate(true);

        QWidget *dummy_widget = new QWidget;
        dummy_widget->setContentsMargins(0,0,0,0);
        dummy_widget->setLayout(vertical_layout);
        setCentralWidget(dummy_widget);


        QRect desktop_rect = QApplication::desktop()->screenGeometry();
        move(settings->value("window_location_x", desktop_rect.width()/4).toInt(),
             settings->value("window_location_y", desktop_rect.height()/4).toInt());
        resize(settings->value("window_size_width", 1).toInt(),
               settings->value("window_size_height", 1).toInt());


        display_throttle.break_time_ms = 40; //25 fps


        broadcaster_thread.start();
        audio_in_thread.start();
        recorder_thread.start();
        take_screenshot();
}

Kumari::~Kumari() {
        if (currently_streaming || currently_recording) stop();

        broadcaster_thread.quit();
        broadcaster_thread.wait();

        audio_in_thread.quit();
        audio_in_thread.wait();

        recorder_thread.quit();
        recorder_thread.wait();

        save_settings();
}

void Kumari::save_settings() {
        settings->setValue("window_location_x", pos().x());
        settings->setValue("window_location_y", pos().y());
        settings->setValue("window_size_width", width());
        settings->setValue("window_size_height", height());

        settings->setValue("x", x_spinbox->value());
        settings->setValue("y", y_spinbox->value());
        settings->setValue("width", width_spinbox->value());
        settings->setValue("height", height_spinbox->value());
        settings->setValue("framerate", framerate_spinbox->value());

        settings->setValue("stream_checkbox_checked", stream_checkbox->isChecked());
        settings->setValue("record_checkbox_checked", record_checkbox->isChecked());

        settings->setValue("video_bitrate_kbit", video_bitrate_spinbox->value());
        settings->setValue("audio_bitrate_kbit", audio_bitrate_spinbox->value());

        settings->setValue("stream_preset_number", stream_slider->value());
        settings->setValue("record_preset_number", record_slider->value());

        settings->setValue("audio_in_device", audio_in_combobox->currentText());
        settings->setValue("audio_out_device", audio_out_combobox->currentText());
        settings->setValue("playthrough", playthrough_checkbox->isChecked());

        settings->sync(); //why is this necessary? (20130207)
}

QSpinBox *Kumari::make_spinbox() {
        QSpinBox *retval = new QSpinBox;
        retval->setRange(0, 99999);
        return retval;
}

QSlider *Kumari::make_slider() {
        QSlider *retval = new QSlider(Qt::Horizontal);
        retval->setRange(0, 9);
        return retval;
}

void Kumari::display_vate(bool boolean) {
        if (boolean) {
                display = new Webcam_Display;
                connect(this, SIGNAL(display_frame(QImage)),
                       display, SLOT(display_image(QImage)), Qt::UniqueConnection);
                vertical_layout->addWidget(display, 1);
        } else {
                disconnect(this, SIGNAL(display_frame(QImage)),
                        display, SLOT(display_image(QImage)));
                vertical_layout->removeWidget(display);
                display->deleteLater();
                adjustSize();
        }
        //display.setShown(boolean); //this totally borks the Webcam_Display object's appearance ... (20130509)
}

void Kumari::start() {
        if (currently_streaming || currently_recording) return;

        if (!stream_checkbox->isChecked() && !record_checkbox->isChecked()) return;

        if (stream_checkbox->isChecked()) {
                if (settings->value("streaming_url").toString().isEmpty()) {
                        if (!set_streaming_url()) {
                                return;
                        }
                }
        }
        if (record_checkbox->isChecked()) {
                if (settings->value("recording_filepath").toString().isEmpty()) {
                        if (!set_recording_filepath()) {
                                return;
                        }
                }
        }

        if (stream_checkbox->isChecked()) {
                emit broadcaster_start(settings->value("streaming_url").toString(),
                                       width_spinbox->value(), height_spinbox->value(),
                                       framerate_spinbox->value(), presets.at(stream_slider->value()),
                                       video_bitrate_spinbox->value(),
                                       audio_bitrate_spinbox->value(), 44100, 2);
                currently_streaming = true;
        }
        if (record_checkbox->isChecked()) {
                QString filename_base = settings->value("recording_filepath").toString();
                filename_base.replace(QRegExp("\\.avi$", Qt::CaseInsensitive), "");
                QString ext = ".avi";
                QString out_filepath = filename_base + ext;
                QFileInfo info(out_filepath);
                uint inc = 1;
                while (info.exists()) {
                        inc++;
                        out_filepath = filename_base + QString("%1").arg(inc) + ext;
                        info.setFile(out_filepath);
                }

                emit recording_start(out_filepath,
                                     width_spinbox->value(), height_spinbox->value(),
                                     framerate_spinbox->value(), presets.at(record_slider->value()),
                                     44100, 2);
                currently_recording = true;
        }

        display_vate(false);
        ui_vate(false);

        start_stop_button->setText(tr("Stop"));
        disconnect(start_stop_button, SIGNAL(clicked()), this, SLOT(start()));
        connect(start_stop_button, SIGNAL(clicked()), this, SLOT(stop()), Qt::UniqueConnection);
}

void Kumari::stop() {
        if (!currently_streaming && !currently_recording) return;

        if (stream_checkbox->isChecked())
                emit broadcaster_stop();

        if (record_checkbox->isChecked())
                emit recording_stop();

        ui_vate(true);

        start_stop_button->setText(tr("Start"));
        disconnect(start_stop_button, SIGNAL(clicked()), this, SLOT(stop()));
        connect(start_stop_button, SIGNAL(clicked()), this, SLOT(start()), Qt::UniqueConnection);

        display_vate(true);

        currently_streaming = currently_recording = false;
}

void Kumari::ui_vate(bool boolean) {
        foreach (QAction *action, actions) {
                action->setEnabled(boolean);
        }
        foreach (QWidget *widget, widgets_to_vate) {
                widget->setEnabled(boolean);
        }
        if (boolean && !playthrough_checkbox->isChecked()) {
                audio_out_combobox->setEnabled(false);
        }
}

void Kumari::error_dialog_slot(QString text) {
        qDebug() << this << "received error:" << text;
}

void Kumari::take_screenshot() {
        QImage image;
        GTimer profile_timer;
        if (profiling) profile_timer.start();

        QScreen *screen = QGuiApplication::primaryScreen();
        QPixmap screenshot = screen->grabWindow(
                                QApplication::desktop()->winId()
                                ,x_spinbox->value()
                                ,y_spinbox->value()
                                ,width_spinbox->value()
                                ,height_spinbox->value()
                                );
        if (profiling) qDebug() << "\n" << profile_timer.restart();
        image = screenshot.toImage(); //format is QImage::Format_ARGB32_Premultiplied on os x 10.8.3 (20131020)
        if (profiling) qDebug() << profile_timer.restart();

        emit frame_ready(image);
        if (profiling) qDebug() << profile_timer.restart();

        if (!display_throttle.busy()) {
                emit display_frame(image);
                display_throttle.done();
                if (profiling) qDebug() << profile_timer.restart();
        }

        ++frames_captured;

        ++frames_captured_for_fps;
        if (frames_captured_for_fps % (int)framerate_spinbox->value() == 0) {
                double fps_label_value = frames_captured_for_fps / (fps_timer.elapsed() / 1000);
                if (fps_label_value) {
                        fps_label->setText(QString("%1").arg(fps_label_value, 0, 'f', 1));
                }
        }
        if (frames_captured_for_fps % (int)framerate_spinbox->value()*5 == 0) {
                frames_captured_for_fps = 0;
                fps_timer.restart();
        }

        double elapsed = session_timer.elapsed();

        if (profiling) qDebug() << profile_timer.restart();

        double time_to_sleep = qMax(0., (frame_time_ms * frames_captured) - elapsed);
        //qDebug() << "sleeping for" << time_to_sleep;
        QTimer::singleShot(time_to_sleep, this, SLOT(take_screenshot()));
}

void Kumari::fps_changed(int fps) {
        frame_time_ms = (1.0/fps) * 1000.0;

//        screenshot_timer.stop();
//        screenshot_timer.setInterval(frame_time_ms);

        frames_captured = 0;
        frames_captured_for_fps = 0;

//        screenshot_timer.start();
        fps_timer.start();
        session_timer.start();
}

bool Kumari::set_streaming_url() {
        bool ok;
        QString text = QInputDialog::getText(this, tr("Set Streaming URL"),
                                             tr("URL:"), QLineEdit::Password,
                                             settings->value("streaming_url").toString(), &ok);
        if (ok && !text.isEmpty()) {
                settings->setValue("streaming_url", text);
                save_settings();
                return true;
        }
        return false;
}

bool Kumari::set_recording_filepath() {
        QString new_filepath = QFileDialog::getSaveFileName(this, tr("Save Recording As"),
                                                            QStandardPaths::writableLocation(QStandardPaths::DesktopLocation),
                                                            tr("AVI files (*.avi)"));
        if (!new_filepath.isEmpty()) {
                settings->setValue("recording_filepath", new_filepath);
                return true;
        }
        return false;
}

void Kumari::about() {
        QMessageBox::about(this, tr("About Kumari"),

                           tr("<i>\"shining, single singing circle-wing extending\"</i>"

                              "<br><br>\n<a href=\"http://forum.speeddemosarchive.com/post/kumari.html\">Kumari</a> version %1 copyright %2 Taiga Software LLC"
                              "<br><br>\nYour use of this software is governed by the terms of the <a href=\"http://www.gnu.org/licenses/gpl.html\">GPL</a> version 2 or, at your option, any later version."
                              "<br><br>\nThis software uses libraries from the <a href=\"http://ffmpeg.org/\">FFmpeg project</a> under the <a href=\"http://www.gnu.org/licenses/gpl.html\">GPL</a>."
                              ).arg(KUMARI_VERSION).arg(QDate::currentDate().year()));
}

void Kumari::populate_audio_devices(QStringList input_device_names, QStringList output_device_names) {
        audio_in_combobox->addItems(input_device_names);
        audio_out_combobox->addItems(output_device_names);

        int audio_in_index = audio_in_combobox->findText(settings->value("audio_in_device").toString());
        if (audio_in_index != -1) {
                audio_in_combobox->setCurrentIndex(audio_in_index);
        }

        int audio_out_index = audio_out_combobox->findText(settings->value("audio_out_device").toString());
        if (audio_out_index != -1) {
                audio_out_combobox->setCurrentIndex(audio_out_index);
        }

        playthrough_checkbox->setChecked(settings->value("playthrough").toBool());

        audio_finished_populating = true;
        return audio_changed();
}

void Kumari::audio_changed() {
        emit audio_start(audio_in_combobox->currentText(), audio_out_combobox->currentText(), playthrough_checkbox->isChecked());
}

void Kumari::audio_device_changed(QString) {
        if (audio_finished_populating) return audio_changed();
}

void Kumari::audio_playthrough_changed(bool boolean) {
        audio_out_combobox->setEnabled(boolean);
        if (audio_finished_populating) return audio_changed();
}
