#ifndef KUMARI_H
#define KUMARI_H

#include <QMainWindow>
#include <QVBoxLayout>
#include <QSpinBox>
#include <QSlider>
#include <QComboBox>
#include <QCheckBox>
#include <QPushButton>
#include <QLabel>
#include <QMenuBar>
#include <QMenu>
#include <QApplication>
#include <QInputDialog>
#include <QLineEdit>
#include <QMessageBox>
#include <QDesktopWidget>
#include <QFileDialog>
#include <QScreen>
#include <QStackedLayout>

#include "rtmp_broadcaster.h"
#include "audio_in.h"
#include "gtimer.h"
#include "webcam_display.h"
#include "throttle.h"
#include "lossless_recorder.h"
#include "level_indicator.h"

class Kumari : public QMainWindow {
        Q_OBJECT
        
public:
        Kumari(QWidget *parent = 0);
        ~Kumari();
        QMutex ffmpeg_mutex;

signals:
        void broadcaster_start(QString,int,int,double,QString,int,int,int,int);
        void broadcaster_stop();
        void display_frame(QImage);
        void frame_ready(QImage);
        void encode_frame(QImage);
        void recording_start(QString,int,int,double,QString,int,int);
        void recording_stop();
        void audio_start(QString,QString,bool);

private slots:
        void error_dialog_slot(QString);
        void start();
        void stop();
        void take_screenshot();
        void fps_changed(int);
        bool set_streaming_url();
        void save_settings();
        bool set_recording_filepath();
        void about();
        void populate_audio_devices(QStringList,QStringList);
        void audio_device_changed(QString);
        void audio_playthrough_changed(bool);
        void audio_changed();

private:
        QSpinBox *make_spinbox();
        QSlider *make_slider();

        RTMP_Broadcaster broadcaster;
        Audio_In audio_in;
        QThread broadcaster_thread, audio_in_thread;

        bool currently_streaming, currently_recording;

        QSpinBox *x_spinbox, *y_spinbox, *width_spinbox, *height_spinbox, *framerate_spinbox,
        *video_bitrate_spinbox, *audio_bitrate_spinbox;
        QCheckBox *playthrough_checkbox;
        QComboBox *audio_in_combobox, *audio_out_combobox;
        QCheckBox *stream_checkbox, *record_checkbox;
        QPushButton *start_stop_button;
        QSlider *stream_slider, *record_slider;

        QList<QAction *> actions;
        QList<QWidget *> widgets_to_vate;

        QLabel *fps_label;
//        QTimer screenshot_timer;
        GTimer session_timer, fps_timer;
        int frames_captured, frames_captured_for_fps;
        Webcam_Display *display;
        bool profiling;
        QSettings *settings;

        QMenuBar *menu_bar;
        QMenu *stream_menu, *record_menu, *help_menu;
        QAction *set_streaming_url_act, *set_recording_filepath_act,
        *about_act, *about_qt_act;
        double frame_time_ms;

        void display_vate(bool);
        void ui_vate(bool);

        QThread recorder_thread;
        Lossless_Recorder lossless_recorder;

        Throttle display_throttle;
        QStringList presets;

        QVBoxLayout *vertical_layout;

        bool audio_finished_populating;

        Level_Indicator level_indicator;

};

#endif // KUMARI_H
